

set -ex



export CONTINUE_THROUGH_ERROR=1
python ./test/run_test.py --core || true
pip check
python -c "import torch; print(torch.__version__)"
python -c "import torch; import numpy"
python -c "import numpy; import torch"
python -c "import torch; assert torch.distributed.is_available()"
exit 0
